__version__ = "unknown"
from ._version import __version__


from mrestimator import utility as ut
ut.initialize()
log = ut.log

from .coefficients import CoefficientResult, coefficients
from .fit          import *
from .input_output import *
from .simulate     import *
from .wrapper      import *




















